﻿// <copyright company="PEER Group GmbH">
//    Copyright (c) PEER Group GmbH. All rights reserved.
// </copyright>

using System;
using System.Collections.Specialized;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using Infragistics.Windows.DockManager;
using Infragistics.Windows.DockManager.Events;
using Microsoft.Practices.Prism.Events;
using Microsoft.Practices.Prism.Regions;

//using PeerGroup.GCL.Logging;

namespace SpreadsheetOwnerErrorSampleV3.Core.Prism
{
	/// <summary>
	/// The region adapter for the TabGroupPane. All changes have been applied as per Update 3 from website.
	/// </summary>
	/// <remarks>
	/// Based on source:
	/// http://www.infragistics.com/community/blogs/blagunas/archive/2012/09/13/xamdockmanager-a-prism-regionadapter.aspx
	/// Update 2: http://brianlagunas.com/xamdockmanageran-updated-prism-region-adapter/
	/// update 3: http://brianlagunas.com/xamdockmanager-prism-region-adapter-update-3/
	/// Written by Brian Lagunas.
	/// Blog: http://brianlagunas.com.
	/// Twitter: @brianlagunas.
	/// Email: blagunas@infragistics.com.
	/// </remarks>
	public class TabGroupPaneRegionAdapter : RegionAdapterBase<TabGroupPane>
	{
		private static readonly DependencyProperty sIsGeneratedProperty =
			DependencyProperty.RegisterAttached("sIsGenerated", typeof(bool), typeof(TabGroupPaneRegionAdapter), null);
		
		private static readonly DependencyProperty sRegionProperty =
			DependencyProperty.RegisterAttached("sRegion", typeof(IRegion), typeof(TabGroupPaneRegionAdapter), null);

		private readonly IEventAggregator mEventAggregator;
		
		public TabGroupPaneRegionAdapter(IRegionBehaviorFactory regionBehaviorFactory, IEventAggregator eventAggregator)
			: base(regionBehaviorFactory) => mEventAggregator = eventAggregator;

		/// <summary> Adapts the specified region. </summary>
		/// <param name="region">The region.</param>
		/// <param name="regionTarget">The region target.</param>
		protected override void Adapt(IRegion region, TabGroupPane regionTarget)
		{
			if (regionTarget.ItemsSource != null)
			{
				throw new InvalidOperationException("ItemsSource property is not empty.");
			}

			SynchronizeItems(region, regionTarget);

			region.Views.CollectionChanged += (sender, e) =>
			{
				OnViewsCollectionChanged(sender, e, region, regionTarget);
			};
		}
		
		protected override void AttachBehaviors(IRegion region, TabGroupPane regionTarget)
		{
			base.AttachBehaviors(region, regionTarget);

			if (!region.Behaviors.ContainsKey(TabGroupPaneRegionActiveAwareBehavior.BEHAVIOR_KEY))
			{
				region.Behaviors.Add(TabGroupPaneRegionActiveAwareBehavior.BEHAVIOR_KEY,
					new TabGroupPaneRegionActiveAwareBehavior
					{
						HostControl = regionTarget, EventAggregator = mEventAggregator
					});
			}

			if (!region.Behaviors.ContainsKey(TabGroupPaneRegionSelectionChangeBehavior.BEHAVIOR_KEY))
			{
				region.Behaviors.Add(TabGroupPaneRegionSelectionChangeBehavior.BEHAVIOR_KEY,
					new TabGroupPaneRegionSelectionChangeBehavior
					{
						HostControl = regionTarget
					});
			}
		}
		
		protected override IRegion CreateRegion()
		{
			return new SingleActiveRegion();
		}
		
		private void ClearContainerForItem(ContentPane contentPane)
		{
			if ((bool)contentPane.GetValue(sIsGeneratedProperty))
			{
				contentPane.ClearValue(ContentPane.HeaderProperty); // Remove any bindings
				(contentPane.DataContext as IDisposable)?.Dispose();
				contentPane.Content = null;
				contentPane.DataContext = null;
			}
		}
		
		private void OnViewsCollectionChanged(object sender, NotifyCollectionChangedEventArgs e, IRegion region,
			TabGroupPane regionTarget)
		{
			if (e.Action == NotifyCollectionChangedAction.Add)
			{
				// We want to add them behind any previous views that may have been manually declare in XAML or injected
				var startIndex = e.NewStartingIndex;
				foreach (var newItem in e.NewItems)
				{
					var contentPane = PrepareContainerForItem(newItem, region);
					if (startIndex > regionTarget.Items.Count)
						startIndex = regionTarget.Items.Count;

					if (startIndex < 0)
						startIndex = 0;
					
					regionTarget.BringIntoView();
					regionTarget.Items.Insert(startIndex++, contentPane);
				}
			}
			else if (e.Action == NotifyCollectionChangedAction.Remove)
			{
				if (regionTarget.Items.Count == 0)
					return;

				var contentPanes = XamDockManager.GetDockManager(regionTarget)
							.GetPanes(PaneNavigationOrder.VisibleOrder);
				foreach (var contentPane in contentPanes)
				{
					if (e.OldItems.Contains(contentPane) || e.OldItems.Contains(contentPane.Content))
						contentPane.ExecuteCommand(ContentPaneCommands.Close);
				}
			}
		}
		
		private void SynchronizeItems(IRegion region, TabGroupPane regionTarget)
		{
			if (regionTarget.Items.Count > 0)
			{
				foreach (var item in regionTarget.Items)
				{
					PrepareContainerForItem(item, region);
					region.Add(item);
				}
			}
		}
		
		private ContentPane PrepareContainerForItem(object item, IRegion region)
		{
			ContentPane container = item as ContentPane;
			if (container == null)
			{
				container = new ContentPane
				{
					Content = item,
					DataContext = ResolveDataContext(item)
				};

				container.AllowDocking = false;
				container.SetValue(sIsGeneratedProperty, true); // We generated this one
				try
				{
					CreateDockAwareBindings(container);
				}
				catch (Exception e)
				{
					ErrorHandler.OnError(e, "Failed to open view container");
				}
			}
			
			container.SetValue(sRegionProperty, region);

			// Make it easy on ourselves and have the pane manage removing itself from the XamDockManager
			container.CloseAction = PaneCloseAction.RemovePane;
			container.Closed += OnContainerClosed;
			container.Closing += OnContainerClosing;

			return container;
		}
		
		private void OnContainerClosed(object sender, PaneClosedEventArgs e)
		{
			var contentPane = sender as ContentPane;
			if (contentPane != null)
			{
				// No memory leaks
				contentPane.Closed -= OnContainerClosed;
				contentPane.Closing -= OnContainerClosing;

				// Get the region associated with the ContentPane so that we can remove it.
				IRegion region = contentPane.GetValue(sRegionProperty) as IRegion;
				if (region != null)
				{
					if (region.Views.Contains(contentPane)) // We are dealing with a ContentPane directly
						region.Remove(contentPane);

					var item = contentPane.Content; // This view was injected and set as the content of our ContentPane
					if (item != null && region.Views.Contains(item))
						region.Remove(item);

					var regionRemovable = GetRegionRemovable(contentPane);
					regionRemovable?.Remove(region.Name);
				}

				ClearContainerForItem(contentPane); // Reduce memory leaks
			}
		}
		
		private void OnContainerClosing(object sender, PaneClosingEventArgs e)
		{
			var contentPane = sender as ContentPane;
			if (contentPane == null)
				return;

			var regionRemovable = GetRegionRemovable(contentPane);
			if (regionRemovable != null)
			{
				// Get the region associated with the ContentPane so that we can remove it.
				var region = contentPane.GetValue(sRegionProperty) as IRegion;
				var isRemoveAllowed = regionRemovable.IsRemoveAllowed(region.Name);

				// Activate the view if closing not allowed (due isDirty check and etc.)
				if (!isRemoveAllowed && !contentPane.IsFocused)
					contentPane.Activate();

				e.Cancel = !isRemoveAllowed;
			}
		}
		
		private IRegionRemovable GetRegionRemovable(ContentPane contentPane)
		{
			var frameworkElement = contentPane?.Content as FrameworkElement;
			if (frameworkElement == null)
				return null;

			return frameworkElement as IRegionRemovable ?? frameworkElement.DataContext as IRegionRemovable;
		}
		
		private void CreateDockAwareBindings(ContentPane contentPane)
		{
			var binding = new Binding("Header");

			var dockAware = ResolveDockAwareImplementation(contentPane);

			binding.Source = dockAware ?? throw new ApplicationException("Unable to open unregistered view");

			// Set the tab header template.
			contentPane.TabHeaderTemplate = CreateHeaderTemplate(dockAware);

			// Set the content pane properties.
			contentPane.AllowClose = dockAware.AllowClose;
			contentPane.AllowFloatingOnly = false; // dockAware.AllowFloatingOnly;
			contentPane.AllowDocking = false; // dockAware.AllowDocking;
			contentPane.AllowInDocumentHost = false; // dockAware.AllowInDocumentHost;
			contentPane.CloseButtonVisibility = Visibility.Visible; //dockAware.CloseButtonVisibility;
			contentPane.IsPinned = dockAware.IsPinned;

			contentPane.SetBinding(ContentPane.HeaderProperty, binding);
			contentPane.IsVisibleChanged += (sender, args) =>
			{
				//var arg = new ViewContentState(contentPane.IsVisible, contentPane.Content, contentPane.DataContext);
				//mEventAggregator.GetEvent<PageVisibilityChangedEvent>().Publish(arg);
			};
		}

		private IDockAware ResolveDockAwareImplementation(ContentPane contentPane)
		{
			if (contentPane == null)
				throw new ArgumentNullException(nameof(contentPane));

			// Let's first check the item that was injected for IDockAware. This may be a View or ViewModel
			if (contentPane.Content is IDockAware content)
				return content;

			// Nothing was found on the item being injected, let's check the DataContext
			if (contentPane.DataContext is IDockAware dataContext)
				return dataContext;
			
			return null;
		}
		
		private DataTemplate CreateHeaderTemplate(IDockAware dockAware)
		{
			const double ELEMENT_HEIGHT = 16.0;

			// Build tab header template.
			var headerTemplate = new DataTemplate();

			var stackPanelFactory = new FrameworkElementFactory(typeof(StackPanel));
			stackPanelFactory.SetValue(StackPanel.OrientationProperty, Orientation.Horizontal);
			stackPanelFactory.SetValue(FrameworkElement.MarginProperty, new Thickness(2, 0, 2, 0));

			var imageFactory = new FrameworkElementFactory(typeof(Image));
			imageFactory.SetValue(FrameworkElement.HeightProperty, ELEMENT_HEIGHT);
			imageFactory.SetValue(FrameworkElement.MarginProperty, new Thickness(0, 0, 2, 0));
			stackPanelFactory.AppendChild(imageFactory);

			var textBlockFactory = new FrameworkElementFactory(typeof(TextBlock));
			var headerBinding = new Binding("Header") { Source = dockAware };
			textBlockFactory.SetBinding(TextBlock.TextProperty, headerBinding);
			textBlockFactory.SetValue(FrameworkElement.MarginProperty, new Thickness(2, 0, 0, 0));
			stackPanelFactory.AppendChild(textBlockFactory);

			headerTemplate.VisualTree = stackPanelFactory;
			return headerTemplate;
		}
		
		private object ResolveDataContext(object item)
		{
			FrameworkElement frameworkElement = item as FrameworkElement;
			return frameworkElement == null ? item : frameworkElement.DataContext;
		}
	}
}
